import os


def test_paths(tmpdir, HOME):
    from homely._utils import isnecessarypath
    l1 = os.path.join(tmpdir, 'link1')
    d1 = os.path.join(tmpdir, 'dir1')
    d1s = os.path.join(d1, 'subdir')
    l1s = os.path.join(l1, 'subdir')

    os.mkdir(d1)
    os.mkdir(d1s)
    os.symlink(d1, l1)
    assert os.path.isdir(l1s)

    assert isnecessarypath(tmpdir, d1)
    assert isnecessarypath(tmpdir, l1)
    assert isnecessarypath(l1, l1s)
    assert isnecessarypath(d1, l1s)
    assert not isnecessarypath(d1, l1)


fixed = [
    'http://www.foo.com/foo.txt',
    'git+ssh://git.example.com/example/bar',
    'git+ssh://foo@example.com/example/bar',
    'homely.test.repo:///foo/something/bar',
]


def test_expansion_homepath(tmpdir):
    from homely._utils import _homepath2real

    # test how various paths expand out to the home dir
    os.environ["HOME"] = tmpdir

    dir1 = os.path.join(tmpdir, 'dir1')
    link1 = tmpdir + "/link1"

    os.mkdir(dir1)
    os.symlink(dir1, link1)

    assert _homepath2real('.vimrc') == tmpdir + '/.vimrc'
    assert _homepath2real('.vimrc') == tmpdir + '/.vimrc'
    assert _homepath2real('bin/ack') == tmpdir + '/bin/ack'
    assert _homepath2real('link1/foo.txt') == tmpdir + '/link1/foo.txt'
    assert _homepath2real('/etc/hosts') == '/etc/hosts'

    assert _homepath2real('~/.vimrc') == tmpdir + '/.vimrc'
    assert _homepath2real('$HOME/.vimrc') == tmpdir + '/.vimrc'
    assert _homepath2real('~/bin/ack') == tmpdir + '/bin/ack'

    # check that relative paths like '.' also work
    assert _homepath2real('..') == '..'
    assert _homepath2real('../foo') == '../foo'

    # these should all raise errors
    badpaths = ['.', 'bin/', '/etc/']
    for path in badpaths:
        try:
            _homepath2real(path)
        except Exception:
            pass
        else:
            raise Exception("Expected an error from %r!" % path)

    for path in fixed:
        assert _homepath2real(path) == path


def test_expansion_repopath(tmpdir):
    from homely._utils import _repopath2real
    os.environ["HOME"] = tmpdir

    rdir = os.path.join(tmpdir, 'repo.git')

    # construct a fake repo
    from homely._vcs.testhandler import Repo
    repo = Repo(rdir, False, False, None)
    assert _repopath2real('foo.txt', repo) == rdir + '/foo.txt'
    assert _repopath2real('a/b/c.txt', repo) == rdir + '/a/b/c.txt'
    assert _repopath2real('/tmp/foo.txt', repo) == os.path.realpath('/tmp/foo.txt')
    # if '~' is included in the path then it will always expand out to $HOME
    assert _repopath2real('~/bin/foo.txt', repo) == tmpdir + '/bin/foo.txt'
    assert _repopath2real('$HOME/bin/foo.txt', repo) == tmpdir + '/bin/foo.txt'
    for path in fixed:
        assert _repopath2real(path, repo) == path



from pytest import HOMELY, TempRepo, contents, getsystemfn


def test_lineinfile_knows_about_ownership(HOME, tmpdir):
    system = getsystemfn(HOME)

    # put the 'AAA' line into my file1.txt
    f1 = HOME + '/file1.txt'
    contents(f1, 'AAA\n')

    # create a fake repo and add it
    tr = TempRepo(tmpdir, 'dotfiles')
    contents(tr.remotepath + '/HOMELY.py',
             """
             from homely.general import lineinfile
             lineinfile('file1.txt', 'AAA')
             lineinfile('file1.txt', 'BBB')
             """)
    system(HOMELY('add') + [tr.url])

    # check that my file1.txt now has both lines
    assert contents(f1) == "AAA\nBBB\n"

    # now remove the repo and do a full update to trigger cleanup
    system(HOMELY('forget') + [tr.url])
    system(HOMELY('update'))

    # run homely update a few more times to confuse it
    system(HOMELY('update'))
    system(HOMELY('update'))

    # check that only the 'BBB' line was removed
    assert contents(f1) == "AAA\n"



from setuptools import find_packages, setup

setup(
    name='homely',
    description=('Automate the installation of your personal config files and'
                 ' favourite tools using Python. https://homely.readthedocs.io/'),
    url='https://homely.readthedocs.io/',
    author='Peter Hodge',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Topic :: Utilities',
    ],
    keywords='dotfiles environment configuration tools utilities automation',
    packages=['homely'] + ['homely.{}'.format(p)
                           for p in find_packages('homely')],
    install_requires=['simplejson', 'click', 'requests', 'python-daemon'],
    entry_points={
        'console_scripts': ['homely=homely._cli:main'],
    },
    # automatic version number using setuptools_scm
    setup_requires=['setuptools_scm'],
    use_scm_version={
        "write_to": 'homely/__init__.py',
    },
)



import os
import sys
import time

from click import UsageError, argument, echo, group, option, version_option

from homely import version
from homely._errors import (ERR_NO_COMMITS, ERR_NOT_A_REPO, JsonError,
                            NotARepo, RepoHasNoCommitsError)
from homely._ui import (PROMPT_ALWAYS, PROMPT_NEVER, addfromremote, note,
                        run_update, setallowpull, setverbose, setwantprompt,
                        warn)
from homely._utils import (FAILFILE, OUTFILE, PAUSEFILE, STATUSCODES, RepoInfo,
                           RepoListConfig, UpdateStatus, getstatus, mkcfgdir,
                           saveconfig)
from homely._vcs import getrepohandler

CMD = os.path.basename(sys.argv[0])


class Fatal(Exception):
    pass


def _globals(command):
    def proxy(verbose, alwaysprompt, neverprompt, **kwargs):
        # handle these global options
        setverbose(verbose)
        if alwaysprompt:
            if neverprompt:
                raise UsageError("--alwaysprompt and --neverprompt options"
                                 " cannot be used together")
            setwantprompt(PROMPT_ALWAYS)
        elif neverprompt:
            setwantprompt(PROMPT_NEVER)
        # pass all other arguments on to the command
        command(**kwargs)
    proxy.__name__ = command.__name__
    proxy.__doc__ = command.__doc__
    proxy = option('-a', '--alwaysprompt', is_flag=True,
                   help="Always prompt the user to answer questions, even"
                   " named questions that they have answered on previous runs"
                   )(proxy)
    proxy = option('-n', '--neverprompt', is_flag=True,
                   help="Never prompt the user to answer questions. Questions"
                   " will be answered automatically using the user's previous"
                   " answer or the `noprompt` value.")(proxy)
    proxy = option('-v', '--verbose', 'verbose', is_flag=True,
                   help="Produce extra output")(proxy)
    return proxy


version_message = (
    "%(prog)s {}, running on python {}.{}.{}"
    .format(version, *sys.version_info[0:3])
)

@group()
@version_option(message=version_message)
def homely():
    """
    Single-command dotfile installation.
    """


@homely.command()
# FIXME: add the ability to install multiple things at once
@argument('repo_path')
@argument('dest_path', required=False)
@_globals
def add(repo_path, dest_path):
    '''
    Registers a git repository with homely so that it will run its `HOMELY.py`
    script on each invocation of `homely update`. `homely add` also immediately
    executes a `homely update` so that the dotfiles are installed straight
    away. If the git repository is hosted online, a local clone will be created
    first.

    REPO_PATH
        A path to a local git repository, or the URL for a git repository
        hosted online. If REPO_PATH is a URL, then it should be in a format
        accepted by `git clone`. If REPO_PATH is a URL, you may also specify
        DEST_PATH.
    DEST_PATH
        If REPO_PATH is a URL, then the local clone will be created at
        DEST_PATH. If DEST_PATH is omitted then the path to the local clone
        will be automatically derived from REPO_PATH.
    '''
    mkcfgdir()
    try:
        repo = getrepohandler(repo_path)
    except NotARepo as err:
        echo("ERROR: {}: {}".format(ERR_NOT_A_REPO, err.repo_path))
        sys.exit(1)

    # if the repo isn't on disk yet, we'll need to make a local clone of it
    if repo.isremote:
        localrepo, needpull = addfromremote(repo, dest_path)
    elif dest_path:
        raise UsageError("DEST_PATH is only for repos hosted online")
    else:
        try:
            repoid = repo.getrepoid()
        except RepoHasNoCommitsError as err:
            echo("ERROR: {}".format(ERR_NO_COMMITS))
            sys.exit(1)
        localrepo = RepoInfo(repo, repoid, None)
        needpull = False

    # if we don't have a local repo, then there is nothing more to do
    if not localrepo:
        return

    # remember this new local repo
    with saveconfig(RepoListConfig()) as cfg:
        cfg.add_repo(localrepo)
    success = run_update([localrepo], pullfirst=needpull, cancleanup=True)
    if not success:
        sys.exit(1)


@homely.command()
@option('--format', '-f',
        help="Format string for the output, which will be passed through"
        " str.format(). You may use the following named replacements:"
        "\n%(repoid)s"
        "\n%(localpath)s"
        "\n%(canonical)s")
@_globals
def repolist(format):
    cfg = RepoListConfig()
    for info in cfg.find_all():
        vars_ = dict(
            repoid=info.repoid,
            localpath=info.localrepo.repo_path,
            canonical=(info.canonicalrepo.repo_path
                       if info.canonicalrepo else '')
        )
        print(format % vars_)


@homely.command()
@argument('identifier', nargs=-1)
@_globals
def forget(identifier):
    '''
    Tells homely to forget about a dotfiles repository that was previously
    added. You can then run `homely update` to have homely perform automatic
    cleanup of anything that was installed by that dotfiles repo.

    REPO
        This should be the path to a local dotfiles repository that has already
        been registered using `homely add`. You may specify multiple REPOs to
        remove at once.
    '''
    errors = False
    for one in identifier:
        cfg = RepoListConfig()
        info = cfg.find_by_any(one, "ilc")
        if not info:
            warn("No repos matching %r" % one)
            errors = True
            continue

        # update the config ...
        note("Removing record of repo [%s] at %s" % (
            info.shortid(), info.localrepo.repo_path))
        with saveconfig(RepoListConfig()) as cfg:
            cfg.remove_repo(info.repoid)

    # if there were errors, then don't try and do an update
    if errors:
        sys.exit(1)


@homely.command()
@argument('identifiers', nargs=-1, metavar="REPO")
@option('--nopull', is_flag=True,
        help="Do not use `git pull` or other things that require internet"
        " access")
@option('--only', '-o', multiple=True,
        help="Only process the named sections (whole names only)")
@_globals
def update(identifiers, nopull, only):
    '''
    Performs a `git pull` in each of the repositories registered with
    `homely add`, runs all of their HOMELY.py scripts, and then performs
    automatic cleanup as necessary.

    REPO
        This should be the path to a local dotfiles repository that has already
        been registered using `homely add`. If you specify one or more `REPO`s
        then only the HOMELY.py scripts from those repositories will be run,
        and automatic cleanup will not be performed (automatic cleanup is only
        possible when homely has done an update of all repositories in one go).
        If you do not specify a REPO, all repositories' HOMELY.py scripts will
        be run.

    The --nopull and --only options are useful when you are working on your
    HOMELY.py script - the --nopull option stops you from wasting time checking
    the internet for the same updates on every run, and the --only option
    allows you to execute only the section you are working on.
    '''
    mkcfgdir()
    setallowpull(not nopull)

    cfg = RepoListConfig()
    if len(identifiers):
        updatedict = {}
        for identifier in identifiers:
            repo = cfg.find_by_any(identifier, "ilc")
            if repo is None:
                hint = "Try running %s add /path/to/this/repo first" % CMD
                raise Fatal("Unrecognised repo %s (%s)" % (identifier, hint))
            updatedict[repo.repoid] = repo
        updatelist = updatedict.values()
        cleanup = len(updatelist) == cfg.repo_count()
    else:
        updatelist = list(cfg.find_all())
        cleanup = True
    success = run_update(updatelist,
                         pullfirst=not nopull,
                         only=only,
                         cancleanup=cleanup)
    if not success:
        sys.exit(1)


@homely.command()
@option('--pause', is_flag=True,
        help="Pause automatic updates. This can be useful while you are"
        " working on your HOMELY.py script")
@option('--unpause', is_flag=True, help="Un-pause automatic updates")
@option('--outfile', is_flag=True,
        help="Prints the _path_ of the file containing the output of the"
        " previous 'homely update' run that was initiated by autoupdate.")
@option('--daemon', is_flag=True,
        help="Starts a 'homely update' daemon process, as long as it hasn't"
        " been run too recently")
@option('--clear', is_flag=True,
        help="Clear any previous update error so that autoupdate can initiate"
        " updates again.")
@_globals
def autoupdate(**kwargs):
    options = ('pause', 'unpause', 'outfile', 'daemon', 'clear')
    action = None
    for name in options:
        if kwargs[name]:
            if action is not None:
                raise UsageError("--%s and --%s options cannot be combined"
                                 % (action, name))
            action = name

    if action is None:
        raise UsageError("Either %s must be used"
                         % (" or ".join("--{}".format(o) for o in options)))

    mkcfgdir()
    if action == "pause":
        with open(PAUSEFILE, 'w'):
            pass
        return

    if action == "unpause":
        if os.path.exists(PAUSEFILE):
            os.unlink(PAUSEFILE)
        return

    if action == "clear":
        if os.path.exists(FAILFILE):
            os.unlink(FAILFILE)
        return

    if action == "outfile":
        print(OUTFILE)
        return

    # is an update necessary?
    assert action == "daemon"

    # check if we're allowed to start an update
    status, mtime, _ = getstatus()
    if status == UpdateStatus.FAILED:
        print("Can't start daemon - previous update failed")
        sys.exit(1)
    if status == UpdateStatus.PAUSED:
        print("Can't start daemon - updates are paused")
        sys.exit(1)
    if status == UpdateStatus.RUNNING:
        print("Can't start daemon - an update is already running")
        sys.exit(1)

    # abort the update if it hasn't been long enough
    interval = 20 * 60 * 60
    if mtime is not None and (time.time() - mtime) < interval:
        print("Can't start daemon - too soon to start another update")
        sys.exit(1)

    assert status in (UpdateStatus.OK,
                      UpdateStatus.NEVER,
                      UpdateStatus.NOCONN)

    oldcwd = os.getcwd()
    import daemon
    with daemon.DaemonContext(), open(OUTFILE, 'w') as f:
        try:
            from homely._ui import setstreams
            setstreams(f, f)

            # we need to chdir back to the old working directory or  imports
            # will be broken!
            if sys.version_info[0] < 3:
                os.chdir(oldcwd)

            cfg = RepoListConfig()
            run_update(list(cfg.find_all()),
                       pullfirst=True,
                       cancleanup=True)
        except Exception:
            import traceback
            f.write(traceback.format_exc())
            raise


@homely.command()
@_globals
def updatestatus():
    """
    Returns an exit code indicating the state of the current or previous
    'homely update' process. The exit code will be one of the following:
      0  ..  No 'homely update' process is running.
      2  ..  A 'homely update' process has never been run.
      3  ..  A 'homely update' process is currently running.
      4  ..  Updates using 'autoupdate' are currently paused.
      5  ..  The most recent update raised Warnings or failed altogether.
      1  ..  (An unexpected error occurred trying to get the status.)
    """
    status = getstatus()[0]
    sys.exit(STATUSCODES[status])


def main():
    try:
        # FIXME: always ensure git is installed first
        homely()
    except (Fatal, JsonError) as err:
        echo("ERROR: %s" % err, err=True)
        sys.exit(1)
    finally:
        try:
            import asyncio
            asyncio.get_event_loop().close()
        except ImportError:
            pass


if __name__ == '__main__':
    main()



import contextlib
import os
import re
import shutil
import subprocess
import sys
import tempfile
from functools import partial
from itertools import chain
from os.path import exists, join

import simplejson

from homely._errors import JsonError
from homely._vcs import Repo, fromdict

try:
    import asyncio
    from homely._asyncioutils import _runasync
except ImportError:
    asyncio = None

try:
    # python3.3+
    from importlib.machinery import SourceFileLoader

    def _loadmodule(name, path):
        return SourceFileLoader(name, path).load_module()
except ImportError:
    # python2
    import imp

    def _loadmodule(name, path):
        return imp.load_source(name, path)

if sys.version_info[0] < 3:
    def opentext(path, mode, *args, **kwargs):
        if 'r' in mode:
            mode = "U" + mode
        return open(path, mode, *args, **kwargs)
else:
    # for python3, we open text files with universal newline support
    opentext = partial(open, newline="")


ROOT = join(os.environ['HOME'], '.homely')
REPO_CONFIG_PATH = join(ROOT, 'repos.json')
ENGINE2_CONFIG_PATH = join(ROOT, 'engine2.json')
FACT_CONFIG_PATH = join(ROOT, 'facts.json')

# contains the PID of the currently running homely process
RUNFILE = join(ROOT, "update-running")
# written to when a complete update is finished successfully
TIMEFILE = join(ROOT, "update-time")
# contains the name of the section currently being executed by 'homely update'
SECTIONFILE = join(ROOT, "update-section")
# this file is touched when a 'homely update' of using all sections is
# unsuccessful
FAILFILE = join(ROOT, "update-failed")
# this file is used to control the pause/unpause state
PAUSEFILE = join(ROOT, "update-paused")
# contains the output of the last 'homely autoupdate' run
OUTFILE = join(ROOT, "autoupdate-output.txt")


_urlregex = re.compile(r"^[a-zA-Z0-9+\-.]{2,20}://")


def mkcfgdir():
    if not exists(ROOT):
        os.mkdir(ROOT)


def _expandpath(path):
    if path.startswith('~'):
        path = os.path.expanduser(path)
    path = os.path.expandvars(path)
    if not (path.startswith('/') or _urlregex.match(path)):
        path = os.path.realpath(path)
    return path


def _repopath2real(path, repo):
    assert isinstance(repo, Repo)
    assert not path.endswith('/')
    assert not repo.isremote
    if path.startswith('~'):
        path = os.path.expanduser(path)
    path = os.path.expandvars(path)
    if not _urlregex.match(path):
        if not path.startswith('/'):
            path = join(repo.repo_path, path)
        path = os.path.realpath(path)
    return path


def _homepath2real(path):
    assert not path.endswith('/')
    # expand ~
    if path.startswith('~'):
        path = os.path.expanduser(path)
    # expand variables
    path = os.path.expandvars(path)
    head = path.split(os.sep, 1)[0]
    # paths relative to the current dir are not allowed
    assert head != '.', 'Relative path not allowed'
    if not (head in ('', '..') or _urlregex.match(path)):
        path = join(os.environ['HOME'], path)
    return path


def run(cmd, stdout=None, stderr=None, **kwargs):
    """
    A blocking wrapper around subprocess.Popen(), but with a simpler interface
    for the stdout/stderr arguments:

    stdout=False / stderr=False
        stdout/stderr will be redirected to /dev/null (or discarded in some
        other suitable manner)
    stdout=True / stderr=True
        stdout/stderr will be captured and returned as a list of lines.
    stdout=None
        stdout will be redirected to the python process's stdout, which may be
        a tty (same as using stdout=subprocess.None)
    stderr=None:
        stderr will be redirected to the python process's stderr, which may be
        a tty (same as using stderr=subprocess.None)
    stderr="STDOUT"
        Same as using stderr=subprocess.STDOUT

    The return value will be a tuple of (exitcode, stdout, stderr)

    If stdout and/or stderr were not captured, they will be None instead.
    """
    devnull = None
    try:
        stdoutfilter = None
        stderrfilter = None

        wantstdout = False
        wantstderr = False
        if stdout is False:
            devnull = open('/dev/null', 'w')
            stdout = devnull
        elif stdout is True:
            stdout = subprocess.PIPE
            wantstdout = True
        elif callable(stdout):
            stdoutfilter = partial(stdout)
            stdout = subprocess.PIPE
        else:
            assert stdout is None, "Invalid stdout %r" % stdout

        if stderr is False:
            if devnull is None:
                devnull = open('/dev/null', 'w')
            stderr = devnull
        elif stderr is True:
            stderr = subprocess.PIPE
            wantstderr = True
        elif stderr == "STDOUT":
            stderr = subprocess.STDOUT
        elif callable(stderr):
            stderrfilter = partial(stderr)
            stderr = subprocess.PIPE
        else:
            assert stderr is None, "Invalid stderr %r" % stderr

        if (stdoutfilter or stderrfilter) and asyncio:
            # run background process asynchronously and filter output as
            # it is running
            exitcode, out, err, = _runasync(stdoutfilter,
                                            stderrfilter,
                                            cmd,
                                            stdout=stdout,
                                            stderr=stderr,
                                            **kwargs)
            if not wantstdout:
                out = None
            if not wantstderr:
                err = None
            return exitcode, out, err

        proc = subprocess.Popen(cmd, stdout=stdout, stderr=stderr, **kwargs)
        out, err = proc.communicate()
        if not wantstdout:
            if stdoutfilter:
                stdoutfilter(out, True)
            out = None
        if not wantstderr:
            if stderrfilter:
                stderrfilter(err, True)
            err = None
        return proc.returncode, out, err
    finally:
        if devnull is not None:
            devnull.close()


def haveexecutable(name):
    exitcode = run(['which', name], stdout=False, stderr=False)[0]
    if exitcode == 0:
        return True
    if exitcode == 1:
        return False
    raise SystemError("Unexpected return value from 'which {}'".format(name))


class JsonConfig(object):
    jsonpath = None
    jsondata = None

    def __init__(self):
        # load up the default json until we know that we can load something from the file
        self.jsondata = self.defaultjson()

        if not os.path.exists(self.jsonpath):
            return
        try:
            with open(self.jsonpath, 'r') as f:
                data = f.read()
                if not len(data):
                    return
                self.jsondata = simplejson.loads(data)
                self.checkjson()
        except simplejson.JSONDecodeError:
            raise JsonError("%s does not contain valid JSON" % self.jsonpath)

    def checkjson(self):
        """
        Child classes should override this method to check if self.jsondata is
        sane.
        """
        raise Exception("This method needs to be overridden")

    def defaultjson(self):
        """
        Child classes should override this method to return the default json
        object for when the config file doesn't exist yet.
        """
        raise Exception("This method needs to be overridden")

    def writejson(self):
        # make dirs needed for config file
        parentdir = os.path.dirname(self.jsonpath)
        if not os.path.exists(parentdir):
            os.makedirs(parentdir, mode=0o755)
        # write the config file now
        dumped = simplejson.dumps(self.jsondata, indent=' ' * 4)
        with open(self.jsonpath, 'w') as f:
            f.write(dumped)


class RepoListConfig(JsonConfig):
    jsonpath = REPO_CONFIG_PATH
    jsondata = None

    def defaultjson(self):
        return []

    def checkjson(self):
        assert isinstance(self.jsondata, list)
        for row in self.jsondata:
            assert 'repoid' in row
            assert 'localrepo' in row
            assert 'localpath' in row
            if 'canonicalpath' in row:
                assert 'canonicalrepo' in row

    def add_repo(self, info):
        assert isinstance(info, RepoInfo)
        modified = False
        for row in self.jsondata:
            if row["repoid"] == info.repoid:
                row.update(self._infotodict(info))
                modified = True
                break
        if not modified:
            self.jsondata.append(self._infotodict(info))

    @staticmethod
    def _infotodict(info):
        assert isinstance(info, RepoInfo)
        ret = {
            "repoid": info.repoid,
            "localpath": info.localrepo.repo_path,
            "localrepo": info.localrepo.asdict(),
        }
        if info.canonicalrepo is not None:
            ret["canonicalpath"] = info.canonicalrepo.repo_path
            ret["canonicalrepo"] = info.canonicalrepo.asdict()
        return ret

    def _infofromdict(self, row):
        localrepo = fromdict(row["localrepo"])
        assert localrepo is not None
        if row.get("canonicalpath"):
            canonical = fromdict(row["canonicalrepo"])
        else:
            canonical = None
        return RepoInfo(
            localrepo,
            row["repoid"],
            canonical,
        )

    def remove_repo(self, repoid):
        newdata = []
        for row in self.jsondata:
            if row["repoid"] != repoid:
                newdata.append(row)
        self.jsondata = newdata

    def find_by_id(self, repoid):
        """
        Returns the repo with the specified <repoid>
        """
        for row in self.jsondata:
            if repoid == row["repoid"]:
                return self._infofromdict(row)

    def find_by_localpath(self, path):
        """
        Returns the repo with the specified local <path>
        """
        # note that the paths in self.jsondata were already _homepath2real()'d
        # in the class' __init__()
        resolved = _homepath2real(path)
        for row in self.jsondata:
            if resolved == os.path.realpath(row["localpath"]):
                return self._infofromdict(row)

    def find_by_canonical(self, repo_path):
        for row in self.jsondata:
            if repo_path == row.get("canonicalpath"):
                return self._infofromdict(row)

    def find_by_any(self, identifier, how):
        """
        how should be a string with any or all of the characters "ilc"
        """
        if "i" in how:
            match = self.find_by_id(identifier)
            if match:
                return match
        if "l" in how:
            match = self.find_by_localpath(identifier)
            if match:
                return match
        if "c" in how:
            match = self.find_by_canonical(identifier)
            if match:
                return match

    def find_all(self):
        for row in self.jsondata:
            yield self._infofromdict(row)

    def repo_count(self):
        return len(self.jsondata)


class RepoScriptConfig(JsonConfig):
    jsondata = None

    def __init__(self, info):
        assert isinstance(info, RepoInfo)
        self.jsonpath = join(ROOT, 'repos', info.repoid + '.json')
        super(RepoScriptConfig, self).__init__()

    @staticmethod
    def remove(info):
        assert isinstance(info, RepoInfo)
        os.unlink(join(ROOT, 'repos', info.repoid + '.json'))

    def defaultjson(self):
        # TODO: prevthings and prevchanges are not needed with the new engine
        return {
            # a list of things that were installed on the last run
            "prevthings": [],
            "prevchanges": {},
            "questions": {},
        }

    def checkjson(self):
        pass

    def getthings(self):
        import homely.general
        import homely.install
        modules = [homely.general, homely.install]
        for thing in self.jsondata['prevthings']:
            class_ = thing["class"]
            identifiers = thing["identifiers"]
            for module in modules:
                if hasattr(module, class_):
                    yield getattr(module, class_).fromidentifiers(identifiers)
                    break
            else:
                raise Exception("No modules own %s" % class_)

    def clearthings(self):
        self.jsondata["prevthings"] = []
        return self.jsondata["prevchanges"]

    @staticmethod
    def _asdict(thing):
        return {"class": thing.__class__.__name__,
                "identifiers": thing.identifiers}

    def addthing(self, thing):
        self.jsondata["prevthings"].append(self._asdict(thing))
        self.jsondata["prevchanges"].setdefault(thing.uniqueid, {})

    def removething(self, thing):
        thingdict = self._asdict(thing)
        prevthings = [t for t in self.jsondata["prevthings"] if t != thingdict]
        assert (len(self.jsondata["prevthings"]) - len(prevthings)) == 1
        del self.jsondata["prevchanges"][thing.uniqueid]

    def setchanges(self, uniqueid, changes):
        self.jsondata["prevchanges"][uniqueid] = changes

    def getprevchanges(self, uniqueid):
        return self.jsondata["prevchanges"].get(uniqueid, {})

    def getquestionanswer(self, name):
        return self.jsondata["questions"].get(name, None)

    def setquestionanswer(self, name, value):
        self.jsondata["questions"][name] = value


class FactConfig(JsonConfig):
    jsonpath = FACT_CONFIG_PATH

    def checkjson(self):
        pass

    def defaultjson(self):
        return {}


@contextlib.contextmanager
def saveconfig(cfg):
    assert isinstance(cfg, JsonConfig)
    yield cfg
    cfg.writejson()


class RepoInfo(object):
    def __init__(self, localrepo, repoid, canonicalrepo=None):
        if localrepo is not None:
            assert isinstance(localrepo, Repo)
            assert not localrepo.isremote
        if canonicalrepo is not None:
            assert isinstance(canonicalrepo, Repo)
            assert canonicalrepo.isremote
            assert canonicalrepo.iscanonical

        self.localrepo = localrepo
        self.canonicalrepo = canonicalrepo
        self.repoid = repoid

    def shortid(self):
        return self.localrepo.shortid(self.repoid)


class NoChangesNeeded(Exception):
    """See filereplacer() for more info"""


@contextlib.contextmanager
def filereplacer(filepath):
    """
    This context manager yields two file pointers:

    origlines:
        a generator that yields lines from the original file
    tmp:
        a file descriptor as if you had used open(tempfile.mkstemp(), 'w')
    NL:
        one of "\n", "\r\n" or "\r" depending on what is encountered first in
        orig. If orig is empty or doesn't exist, then "\n" is used.

    Note that upon successful exiting of the context manager, the file
    nominated by filepath will be replaced by the temp file. This will be done
    using file renames, so it is as close to atomic as we can get it.

    If the context block raises an exception, the original file is not changed,
    and the temp file is deleted.

    If the context block raises a NoChangesNeeded exception, then any changes
    to tmpfile are discarded.
    """
    # create the tmp dir if it doesn't exist yet
    tmpdir = join(ROOT, 'tmp')
    if not os.path.exists(tmpdir):
        os.makedirs(tmpdir, mode=0o700)
    tmpname = join(tmpdir, os.path.basename(filepath))
    try:
        if exists(filepath):
            shutil.copy2(filepath, tmpname)
        with opentext(tmpname, 'w') as tmp:
            if os.path.exists(filepath):
                with opentext(filepath, 'r') as orig:
                    NL = "\n"
                    origlines = []
                    firstline = None
                    for firstline in orig:
                        break
                    if firstline is not None:
                        stripped = firstline.rstrip('\r\n')
                        NL = firstline[len(stripped):]
                        assert NL in ("\r", "\n", "\r\n"), "Bad NL %r" % NL
                        origlines = chain([stripped],
                                          (l.rstrip('\r\n') for l in orig))
                    yield tmp, origlines, NL
            else:
                yield tmp, None, "\n"
    except NoChangesNeeded:
        if os.path.exists(tmpname):
            os.unlink(tmpname)
    except:
        if os.path.exists(tmpname):
            os.unlink(tmpname)
        raise
    if os.path.exists(filepath):
        os.unlink(filepath)
    os.rename(tmpname, filepath)


def isnecessarypath(parent, child):
    """
    returns True if the file, directory or symlink <parent> is required to
    exist in order for the path <child> to refer to a valid filesystem entry.

    Examples:

    If <parent> refers to a file, and <child> refers to the same file, then
    <parent> must exist in order for <child> to be valid.

    If <parent> refers to a directory, and <child> refers to the same directory
    or anything under that directory, then <parent> must exist in order for
    <child> to be valid.

    If <parent> is a symlink to a directory, and <child> refers to something in
    that directory *and contains the symlink <parent> in its path*, then
    <parent> must continue to exist in order for <child> to be valid.
    """
    assert parent.startswith('/')
    assert child.startswith('/')
    # resolve all symlinks in the parent (except for the final part itself)
    head, tail = os.path.split(parent)
    # expand the head part out to its real path
    head = os.path.realpath(head)
    fullparent = os.path.realpath(parent)
    assert len(tail), "Can't use isancestor() on path ending in /: %s" % parent
    prefix = '/'
    parts = child.split('/')
    while len(parts):
        prefix = os.path.realpath(join(prefix, parts.pop(0)))
        common = os.path.commonprefix([prefix, head])

        # if at any time we stumble upon the parent as we are reconstructing
        # the path, then we are dependent on the parent
        if prefix == fullparent and len(parts):
            return True

        # if they refer to the same thing up to this point, check to see if the
        # next parts are also the same
        if len(common) == len(head):
            # if the next item of child's path is the tail of parent, then they
            # must refer to the same thing
            if len(parts) and tail == parts[0]:
                return True

    return False


@contextlib.contextmanager
def tmpdir(name):
    assert '/' not in name, "Invalid name %r" % name
    tmp = None
    try:
        tmp = tempfile.mkdtemp()
        yield join(tmp, name)
    finally:
        if tmp and exists(tmp):
            shutil.rmtree(tmp)


class UpdateStatus(object):
    OK = "ok"
    NEVER = "never"
    RUNNING = "running"
    FAILED = "failed"
    NOCONN = "noconn"
    DIRTY = "dirty"
    PAUSED = "paused"


STATUSCODES = {
    UpdateStatus.OK: 0,
    UpdateStatus.NEVER: 2,
    UpdateStatus.RUNNING: 3,
    UpdateStatus.FAILED: 4,
    UpdateStatus.NOCONN: 5,
    UpdateStatus.DIRTY: 6,
    UpdateStatus.PAUSED: 7,
}


def getstatus():
    """Get the status of the previous 'homely update', or any 'homely update'
    that may be running in another process.
    """
    if exists(RUNFILE):
        mtime = os.stat(RUNFILE).st_mtime
        with open(SECTIONFILE) as f:
            section = f.read().strip()
        # what section?
        return UpdateStatus.RUNNING, mtime, section
    if exists(PAUSEFILE):
        return UpdateStatus.PAUSED, None, None

    mtime = None
    if exists(TIMEFILE):
        mtime = os.stat(TIMEFILE).st_mtime

    if exists(FAILFILE):
        if not mtime:
            mtime = os.stat(FAILFILE).st_mtime
        # TODO: return a different error code when the error was inability to
        # contact one or more remote servers
        with open(FAILFILE) as f:
            content = f.read().strip()
            if content == UpdateStatus.NOCONN:
                return UpdateStatus.NOCONN, mtime, None
            elif content == UpdateStatus.DIRTY:
                return UpdateStatus.DIRTY, mtime, None
        return UpdateStatus.FAILED, mtime, None

    if mtime is None:
        return UpdateStatus.NEVER, None, None

    return UpdateStatus.OK, mtime, None
